{
    "volume_type_id": "%(uuid)s",
    "control_location": "front-end",
    "deleted": false,
    "created_at": "%(strtime)s",
    "updated_at": null,
    "encryption_id": "%(uuid)s",
    "key_size": 128,
    "provider": "luks",
    "deleted_at": null,
    "cipher": "aes-xts-plain64"
}
