.. -*- rst -*-

API version details
===================

Show API v2 details
~~~~~~~~~~~~~~~~~~~

.. rest_method::  GET /v2/

Shows details for Block Storage API v2.


Response codes
--------------

.. rest_status_code:: success ../status.yaml

   - 200


Request
-------



Response Parameters
-------------------

.. rest_parameters:: parameters.yaml

   - location: location



Response Example
----------------

.. literalinclude:: ./samples/version-show-response.json
   :language: javascript


