{
    "snapshot": {
        "created_at": "2018-09-26T03:45:03.893592",
        "description": "this is a new snapshot",
        "id": "b6314a71-9d3d-439a-861d-b790def0d693",
        "metadata": {
            "manage-snap-meta1": "value1",
            "manage-snap-meta2": "value2",
            "manage-snap-meta3": "value3"
        },
        "name": "new_snapshot",
        "size": 1,
        "status": "creating",
        "updated_at": "null",
        "volume_id": "1df34919-aba7-4a1b-a614-3b409d71ac03"
    }
}