.. _cinder-vmware:

.. list-table:: Description of VMware configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``vmware_adapter_type`` = ``lsiLogic``
     - (String) Default adapter type to be used for attaching volumes.
   * - ``vmware_api_retry_count`` = ``10``
     - (Integer) Number of times VMware vCenter server API must be retried upon connection related issues.
   * - ``vmware_ca_file`` = ``None``
     - (String) CA bundle file to use in verifying the vCenter server certificate.
   * - ``vmware_cluster_name`` = ``None``
     - (Multi-valued) Name of a vCenter compute cluster where volumes should be created.
   * - ``vmware_connection_pool_size`` = ``10``
     - (Integer) Maximum number of connections in http connection pool.
   * - ``vmware_host_ip`` = ``None``
     - (String) IP address for connecting to VMware vCenter server.
   * - ``vmware_host_password`` = ``None``
     - (String) Password for authenticating with VMware vCenter server.
   * - ``vmware_host_port`` = ``443``
     - (Port number) Port number for connecting to VMware vCenter server.
   * - ``vmware_host_username`` = ``None``
     - (String) Username for authenticating with VMware vCenter server.
   * - ``vmware_host_version`` = ``None``
     - (String) Optional string specifying the VMware vCenter server version. The driver attempts to retrieve the version from VMware vCenter server. Set this configuration only if you want to override the vCenter server version.
   * - ``vmware_image_transfer_timeout_secs`` = ``7200``
     - (Integer) Timeout in seconds for VMDK volume transfer between Cinder and Glance.
   * - ``vmware_insecure`` = ``False``
     - (Boolean) If true, the vCenter server certificate is not verified. If false, then the default CA truststore is used for verification. This option is ignored if "vmware_ca_file" is set.
   * - ``vmware_max_objects_retrieval`` = ``100``
     - (Integer) Max number of objects to be retrieved per batch. Query results will be obtained in batches from the server and not in one shot. Server may still limit the count to something less than the configured value.
   * - ``vmware_task_poll_interval`` = ``2.0``
     - (Floating point) The interval (in seconds) for polling remote tasks invoked on VMware vCenter server.
   * - ``vmware_tmp_dir`` = ``/tmp``
     - (String) Directory where virtual disks are stored during volume backup and restore.
   * - ``vmware_volume_folder`` = ``Volumes``
     - (String) Name of the vCenter inventory folder that will contain Cinder volumes. This folder will be created under "OpenStack/<project_folder>", where project_folder is of format "Project (<volume_project_id>)".
   * - ``vmware_wsdl_location`` = ``None``
     - (String) Optional VIM service WSDL Location e.g http://<server>/vimService.wsdl. Optional over-ride to default location for bug work-arounds.
