# Copyright 2013 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_utils import importutils


compute_opts = [
    cfg.StrOpt('compute_api_class',
               default='cinder.compute.nova.API',
               help='The full class name of the '
               'compute API class to use'),
]

CONF = cfg.CONF
CONF.register_opts(compute_opts)


def API():
    compute_api_class = CONF.compute_api_class
    cls = importutils.import_class(compute_api_class)
    return cls()
