# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import cinder.api.common


class ViewBuilder(cinder.api.common.ViewBuilder):
    """Map cinder.volumes.api list_availability_zones response into dicts."""

    def list(self, request, availability_zones):
        def fmt(az):
            return {
                'zoneName': az['name'],
                'zoneState': {'available': az['available']},
            }

        return {'availabilityZoneInfo': [fmt(az) for az in availability_zones]}
